unit FFMS2;

interface

uses
  SysUtils, Classes, Windows, Forms, Graphics,
  Libffms2;

const
  BitmapPixelFormat='bgr24';

type
  TFFmpegSource = class
  private
    FPVideoSource: PFFMS_VideoSource;
    FPVideoInfo: PFFMS_VideoProperties;
    FPIndex: PFFMS_Index;
    FPIndexer: PFFMS_Indexer;
    FPVideoTrack: PFFMS_Track;
    FPTimeBase: PFFMS_TrackTimeBase;
    FTimeStamp: int64;
    FFPS:double;
    FWidth, FHeight: integer;
    FVideoWidth, FVideoHeight: integer;
    FPixelFormat: string;
    FFrameIndex, FReadIndex: integer;
    FPFrame: PFFMS_Frame;

    FErrorMessage: array[0..1023] of ansichar;
    FErrorInfo: TFFMS_ErrorInfo;

    FOnError: TNotifyEvent;

    FIsValid:boolean;

    function GetFirstTrackOfTypeI(indexer:PFFMS_Indexer; traceType:integer):integer;
    function UpsideDown(pData: PByte; linesize, lines: integer): PByte;
    procedure HandleError;
    procedure Reset;

    procedure SetWidth(value: integer);
    procedure SetHeight(value: integer);
    procedure SetPixelFormat(value: string);
    function GetFrameCount:integer;
    function GetError: string;
    function GetTimebase: double;
    function GetTime: double;
    function GetDuration: double;

  public
    { Public declarations }
    constructor Create;
    destructor Destroy; override;

    function LoadVideo(fileName:string):boolean;
    function DoIndexing:boolean;
    function SeekFrame(index:integer):boolean;
    function ReadBGRData:PByte;
    function GetFrameBmp(var bmp: TBitmap; index: integer): boolean;

    property Error: string read GetError;
    property IsValid: boolean read FIsValid;
//    property TimeStamp: int64 read FTimeStamp;
//    property TimeBase: double read GetTimebase;
    property Time: double read GetTime;
    property Duration: double read GetDuration;
    property FPS: double read FFPS;
    property FrameCount: integer read GetFrameCount;
    property VideoWidth: integer read FVideoWidth;
    property VideoHeight: integer read FVideoHeight;
    property Width: integer read FWidth write SetWidth;
    property Height: integer read FHeight write SetHeight;
    property PixelFormat: string read FPixelFormat write SetPixelFormat;
    property FrameIndex: integer read FFrameIndex;
  end;

implementation

constructor TFFmpegSource.Create;
begin
  inherited Create;

  FErrorInfo.Buffer:= @FErrorMessage;
  FErrorInfo.BufferSize:= sizeof(FErrorMessage);
  FErrorInfo.ErrorType:= FFMS_ERROR_SUCCESS;
  FErrorInfo.SubType:= FFMS_ERROR_SUCCESS;

  FVideoWidth:= 16;
  FVideoHeight:= 9;
  FWidth:= FVideoWidth;
  FHeight:= FVideoHeight;
  
  Reset;

  FFPS:= 24;
  FIsValid:= false;

  FFMS_Init(0,0);
end;

destructor TFFmpegSource.Destroy;
begin
  FFMS_DestroyVideoSource(FPVideoSource);
end;

procedure TFFmpegSource.HandleError;
begin
  if FErrorInfo.ErrorType<>FFMS_ERROR_SUCCESS then
  begin
    if Assigned(FOnError) then
      FOnError(self);
  end;
end;

procedure TFFmpegSource.SetWidth(value: integer);
begin
  if (value>5) and (value<Screen.width) and (FWidth<>value) then
    FWidth:= value;
end;

procedure TFFmpegSource.SetHeight(value: integer);
begin
  if (value>5) and (value<Screen.Height) and (FHeight<>value) then
    FHeight:= value;
end;

procedure TFFmpegSource.SetPixelFormat(value: string);
var
  pixfmts:array[0..1] of integer;
begin
  if (length(value)<=0) then
    FPixelFormat:= 'none';
  value:= lowercase(value);  
  if (PixelFormat<>value) then
  begin
    FPixelFormat:= value;
  end;
  pixfmts[0] := FFMS_GetPixFmt(pchar(FPixelFormat));
  pixfmts[1] := -1;
  FFMS_SetOutputFormatV2(FPVideoSource, @pixfmts, FWidth, FHeight,
                               FFMS_RESIZER_BICUBIC, @FErrorInfo);
end;

function TFFmpegSource.GetFirstTrackOfTypeI(indexer:PFFMS_Indexer; traceType:integer):integer;
var
  i, trackCount:integer;
begin
	trackCount:= FFMS_GetNumTracksI(Indexer);
  result:= -1;
	for i:=0 to trackCount-1 do
  begin
		if (FFMS_GetTrackTypeI(Indexer, i) = traceType) then
		begin
      result:= i;
      exit;
    end;
	end;
end;


function TFFmpegSource.GetError;

begin

  result:='';

  if FErrorInfo.ErrorType<>FFMS_ERROR_SUCCESS then

    result:= FErrorInfo.Buffer;

end;


function TFFmpegSource.GetTimebase: double;

begin

  if (FPTimeBase=nil) then

    result:= 0.001

  else

    result:= FPTimeBase.Num / FPTimeBase.Den;

end;


function TFFmpegSource.GetTime: double;

begin

  if (FPTimeBase=nil) then

    result:= 0.0

  else

    result:= FTimeStamp * FPTimeBase.Num / FPTimeBase.Den / 1000 - FPVideoInfo.FirstTime;

end;

function TFFmpegSource.GetFrameCount:integer;
begin
  result:= FPVideoInfo.NumFrames;
end;

function TFFmpegSource.GetDuration: double;
begin
  result:= FPVideoInfo.LastTime - FPVideoInfo.FirstTime;
end;

function TFFmpegSource.DoIndexing:boolean;
begin
  FPIndex:= FFMS_DoIndexing2(FPIndexer, 0, @FErrorInfo);
  result:= FPIndex<>nil;
end;

procedure TFFmpegSource.Reset;
begin
  FFrameIndex:= -1;
  FTimeStamp:= 0;
end;

function TFFmpegSource.SeekFrame(index:integer):boolean;
var
  pFrameInfo: PFFMS_FrameInfo;
begin
  if (index=FFrameIndex) then
  begin
    result:=true;
    exit;
  end;
  FPFrame:= FFMS_GetFrame(FPVideoSource, index, @FErrorInfo);
  if (FPFrame=nil) then
  begin
    Reset;
    result:= false;
  end
  else
  begin
    FFrameIndex:= index;
    pFrameInfo:= FFMS_GetFrameInfo(FPVideoTrack, FFrameIndex);
    FTimeStamp:= pFrameInfo.PTS;
    UpsideDown(FPFrame.Data[0], FPFrame.linesize[0], FPFrame.ScaledHeight);
    result:= true;
  end;
end;

function TFFmpegSource.UpsideDown(pData: PByte; linesize, lines: integer): PByte;
var
  p1, p2: PByte;
  tmp: byte;
  x, y: integer;
begin
  result:= pData;
  if (lines<2) or (linesize<3) then
    exit;
  p1:= pData;
  p2:= pData;
  inc(p2, (lines-1) * linesize);

  for y:= 1 to lines div 2 do
  begin
    for x:=1 to linesize do
    begin
      tmp:= p1^;
      p1^:= p2^;
      p2^:= tmp;
      inc(p1);
      inc(p2);
    end;
    dec(p2, linesize * 2);
  end;
end;

function TFFmpegSource.ReadBGRData: PByte;
begin
  result:= FPFrame.Data[0];
end;

function TFFmpegSource.GetFrameBmp(var bmp: TBitmap; index: integer): boolean;
var
  bfh:BITMAPFILEHEADER;
  bmpinfo:BITMAPINFOHEADER;
  size, linesize, y:integer;
  stream: TMemoryStream;
  pFrame: PFFMS_Frame;
begin
  result:= false;
  if (bmp=nil) or (index<0) or (index>=FrameCount-1) then
    exit;
  if (FPixelFormat<>BitmapPixelFormat) then
  begin
    SetPixelFormat(BitmapPixelFormat);
  end;

  if (index<>FFrameIndex) then
  begin
    pFrame:= FFMS_GetFrame(FPVideoSource, index, @FErrorInfo);
    UpsideDown(FPFrame.Data[0], FPFrame.linesize[0], FPFrame.ScaledHeight);
  end
  else
  begin
    pFrame:= FPFrame;
  end;
  if (pFrame.Linesize[0])<=0 then
    exit;
  linesize:= pFrame.Linesize[0];
  size:= FHeight*linesize;
  
  bfh.bfType:= $4d42;
  bfh.bfReserved1:= 0;
  bfh.bfReserved2:= 0;
  bfh.bfOffBits:= sizeof(TBitmapFileHeader) + sizeof(TBitmapInfoHeader);
  bfh.bfSize:= bfh.bfOffBits + size;

  bmpinfo.biSize := sizeof(TBitmapInfoHeader);
  bmpinfo.biWidth := Fwidth;
  bmpinfo.biHeight := Fheight;
  bmpinfo.biPlanes := 1;
  bmpinfo.biBitCount := 24;
  bmpinfo.biCompression := BI_RGB;
  bmpinfo.biSizeImage := 0;
  bmpinfo.biXPelsPerMeter := 0;
  bmpinfo.biYPelsPerMeter := 0;
  bmpinfo.biClrUsed := 0;
  bmpinfo.biClrImportant := 0;

  stream:= TMemoryStream.Create;
  try
    stream.Write(bfh, sizeof(bfh));
    stream.Write(bmpinfo, sizeof(bmpinfo));
    stream.Write(pFrame.Data[0]^, size);
    stream.seek(0, soFromBeginning);
    bmp.LoadFromStream(stream);
  finally
    stream.Free;
  end;
end;

function TFFmpegSource.LoadVideo(fileName:string):boolean;
var
  trackno:integer;
  pFrame: PFFMS_Frame;
  pTrack: PFFMS_Track;
  idxFileName:string;
label
  HASERROR;
begin
  FIsValid:= false;
  result:= false;

  FPIndexer := FFMS_CreateIndexer(pchar(fileName), @FErrorInfo);
  if (FPIndexer=nil) then
  begin
    // handle error (print errinfo.Buffer somewhere)
    HandleError;
    exit;
  end;

  idxFileName:= fileName+'.ffmsindex';
  FPIndex:= FFMS_ReadIndex(pchar(idxFileName), @FErrorInfo);
  if (FPIndex<>nil) and (FFMS_IndexBelongsToFile(FPIndex, pchar(fileName), @FErrorInfo)<>0) then
  begin
    FFMS_DestroyIndex(FPIndex);
    FPIndex:= nil;
  end;
  trackno:= GetFirstTrackOfTypeI(FPIndexer,FFMS_TYPE_VIDEO);
  if (FPIndex<>nil) and (trackNO>=0) then
  begin
    pTrack:= FFMS_GetTrackFromIndex(FPIndex, trackno);
		if (FFMS_GetNumFrames(pTrack) <= 0) then
    begin
      FFMS_DestroyIndex(FPIndex);
			FPIndex:= nil;
    end;
  end;
  if (FPIndex=nil) then
  begin
    FPIndex:= FFMS_DoIndexing2(FPIndexer, FFMS_IEH_ABORT, @FErrorInfo);
    if (FPIndex=nil) then
    begin
      // handle error (you should know what to do by now)
      goto HASERROR;
    end;
    if (FFMS_WriteIndex(pchar(idxFileName), FPIndex, @FErrorInfo)<>0) then
    begin
      goto HASERROR;
    end;
  end;

  if (trackno < 0) then
  begin
    // Retrieve the track number of the first video track
    trackno:= FFMS_GetFirstTrackOfType(FPIndex, FFMS_TYPE_VIDEO, @FErrorInfo);
    if (trackno < 0) then
    begin
      // no video tracks found in the file, this is bad and you should handle it
      goto HASERROR;
    end;
  end;

  // We now have enough information to create the video source object
  FPVideoSource:= FFMS_CreateVideoSource(pchar(fileName), trackno,
                                    FPIndex, 1, FFMS_SEEK_NORMAL, @FErrorInfo);
  if (FPVideoSource = nil) then
  begin
    goto HASERROR;
  end;
  // Since the index is copied into the video source object upon its creation,
  //  we can and should now destroy the index object.
  FFMS_DestroyIndex(FPIndex);

  // Retrieve video properties so we know what we're getting.
  // As the lack of the errmsg parameter indicates, this function cannot fail.
  FPVideoInfo:= FFMS_GetVideoProperties(FPVideoSource);

  // Get the first frame for examination so we know what we're getting. This is required
  // because resolution and colorspace is a per frame property and NOT global for the video.
  pframe:= FFMS_GetFrame(FPVideoSource, 0, @FErrorInfo);;
  if (pframe=nil) then
  begin
    goto HASERROR;
  end;

  // OK!
  FVideoWidth:= pframe.EncodedWidth;
	FVideoHeight:= pframe.EncodedHeight;
  FWidth:= FVideoWidth;
  FHeight:= FVideoHeight;
  
  FFPS:= FPVideoInfo.FPSNumerator / FPVideoInfo.FPSDenominator;

  SetPixelFormat(BitmapPixelFormat);
  if (FErrorInfo.ErrorType>FFMS_ERROR_SUCCESS) then
  begin
    goto HASERROR;
  end;

  FPVideoTrack:= FFMS_GetTrackFromVideo(FPVideoSource);
	if (FPVideoTrack=nil) then
  begin
		goto HASERROR;
  end;

	FPTimeBase:= FFMS_GetTimeBase(FPVideoTrack);
	if (FPTimeBase=nil) then
  begin
    goto HASERROR;
  end;

  {
  	// build list of keyframes and timecodes
	std::vector<int> TimecodesVector;
	for (int CurFrameNum = 0; CurFrameNum < VideoInfo->NumFrames; CurFrameNum++)
  begin
		const FFMS_FrameInfo *CurFrameData = FFMS_GetFrameInfo(FrameData, CurFrameNum);
		if (!CurFrameData)
			throw VideoOpenError("Couldn't get info about frame " + std::to_string(CurFrameNum));

		// keyframe?
		if (CurFrameData->KeyFrame)
			KeyFramesList.push_back(CurFrameNum);

		// calculate timestamp and add to timecodes vector
		int Timestamp = (int)((CurFrameData->PTS * TimeBase->Num) / TimeBase->Den);
		TimecodesVector.push_back(Timestamp);
	end;
	if (TimecodesVector.size() < 2) then
		Timecodes = 25.0;
	else
		Timecodes = agi::vfr::Framerate(TimecodesVector);
  }
  Reset;

  FIsValid:=true;
  result:= FIsValid;

HASERROR:
  HandleError;
end;

end.
