unit ffmpeg;

interface

uses
  SysUtils, Windows, Classes, Controls, Messages,Forms, Graphics, ExtCtrls,
  MMTimer, VFW, GIFImage,
  DXRectTracker, FFMS2;

type
  TFFMpegPlayer = class;

  TPlaybackThread = class(TThread)
  private
    FPlayer: TFFMpegPlayer;
  protected
    procedure Execute; override;
    constructor Create(Player: TFFMpegPlayer);
  end;

  TPlayState=(StateNone, StateLoading, StateLoaded, StatePlaying, StatePlayingAt,
             StatePaused, StateError);

  TSelectedRect = packed record
    Left, Top, Width, Height: integer
  end;

  TFFMpegStartIndexEvent = procedure(Sender: TObject; ExpectedCount: integer) of object;
  TFFMpegIndexChangedEvent = procedure(Sender: TObject; index: integer) of object;

  TFFMpegPlayer = class
  private
    { Private declarations }
    FState: TPlayState;
    FFileName: string;
    FWinControl: TWinControl;
//    FParent: TWinControl;
    FImage: TImage;
    FRectTracker: TDXRectTracker;
    Fhdc: HDC;
    Fhdd: THandle;
    FPlayThread: TPlaybackThread;
    FTimer: TMMTimer;
    FHasError: boolean;
    FBmpInfo: TBitmapInfo;
    FGif: TGifImage;
    FFFMS: TFFmpegSource;
    FPBGRData: PByte;
    FEndFrameIndex: integer;

    FOnStartIndex: TFFMpegStartIndexEvent;
    FOnEndIndex: TNotifyEvent;
    FOnIndexChanged: TFFMpegIndexChangedEvent;
    FOnLoaded: TNotifyEvent;
    FOnPaused: TNotifyEvent;
    FOnRefresh: TNotifyEvent;
    FOnError: TNotifyEvent;
    FOnSelectionResize: TNotifyEvent;

    procedure OnTimer;
    function ReadFrame: boolean;
    function GetError: string;
    function GetTime: double;
    function GetDuration: double;
    function GetTimeStamp: int64;
    function GetFrameCount: integer;
    function GetVideoWidth: integer;
    function GetVideoHeight: integer;
    function GetFPS: double;
    function GetFrameIndex: integer;
    function GetSelection: TSelectedRect;

    procedure CheckIndex(var startIndex: integer; var endIndex: integer);
    procedure SetBitmapHeader;
    procedure SetFileName(value: string);

    procedure SetOnStartIndex(value: TFFMpegStartIndexEvent);
    procedure SetOnEndIndex(value: TNotifyEvent);
    procedure SetOnIndexChanged(value: TFFMpegIndexChangedEvent);
    procedure SetOnLoaded(value: TNotifyEvent);
    procedure SetOnPaused(value: TNotifyEvent);
    procedure SetOnRefresh(value: TNotifyEvent);
    procedure SetOnError(value: TNotifyEvent);
    procedure SetOnSelectionResize(value: TNotifyEvent);

    procedure OpenFile;
    procedure ShowRect;
    procedure HideRect;
    procedure DoError;

  protected

  public
    { Public declarations }
    constructor Create(AOwner: TWinControl);
    destructor Destroy; override;

    function GetIndexTime(index:integer):double;
    function ToGif(startIndex,endIndex,width, height, fps:integer):boolean;

    procedure Play;
    procedure PlayAt(startIndex, endIndex:integer);
    procedure Pause;
    procedure Paint;
    procedure GoNextFrame(count:integer);
    procedure GoPrevFrame(count:integer);
    procedure GoFrame(index:integer);

    property State: TPlayState read fState;
    property FileName:string read fFileName write SetFileName;
    property HasError:boolean read FHasError;
    property ErrorMessage:string read GetError;
    property Duration: double read GetDuration;
    property Time: double read GetTime;
    property TimeStamp: int64 read GetTimeStamp;
    property FrameIndex: integer read GetFrameIndex;
//    property TotalTimeStamp: int64 read FTotalTimeStamp;
    property VideoWidth: integer read GetVideoWidth;
    property VideoHeight: integer read GetVideoHeight;
    property VideoFPS: Double read GetFPS;
//    property FrameIndex: integer read FFrameIndex;
    property FrameCount: integer read GetFrameCount;
    property GifImage: TGifImage read FGif;
    property Selection: TSelectedRect read GetSelection;

    property OnError: TNotifyEvent read FOnError write SetOnError;
    property OnPaused: TNotifyEvent read FOnPaused write SetOnPaused;
    property OnLoaded: TNotifyEvent read FOnLoaded write SetOnLoaded;
    property OnRefresh: TNotifyEvent read FOnRefresh write SetOnRefresh;
    property OnStartIndex: TFFMpegStartIndexEvent read FOnStartIndex write SetOnStartIndex;
    property OnEndIndex: TNotifyEvent read FOnEndIndex write SetOnEndIndex;
    property OnIndexChanged: TFFMpegIndexChangedEvent read FOnIndexChanged write SetOnIndexChanged;

    property OnSelectionResize: TNotifyEvent read FOnSelectionResize write SetOnSelectionResize;
  end;

implementation
{
procedure LoadBmp(bmpFile: String; var bmpinfo: TBitmapInfo; var pBmpData:Pointer);
var
  bmf: TBitmapFileHeader;
  imageSize: LongWord;
  Stream: TFileStream;
begin
   try
    Stream:= TFileStream.Create(bmpFile, fmOpenRead or fmShareDenyWrite);
    Stream.Read(bmf, sizeof(Bmf));
    Stream.Read(bmpinfo, sizeof(bmpinfo));
    imageSize:= bmf.bfSize-bmf.bfOffBits;
    stream.Seek(bmf.bfOffBits,0);

    FreeMemory(pBmpData);
    pBmpData:= GetMemory(imageSize);

    Stream.Read(pBmpData^, ImageSize);
  finally
    FreeAndNil(Stream);
  end;
end;
}

constructor TPlaybackThread.Create(Player: TFFMpegPlayer);
begin
  inherited Create(true);
  FPlayer:= Player;
  FreeOnTerminate:=True;
end;

procedure TPlaybackThread.Execute;
begin
  while not terminated do
  begin
    if (FPlayer.ReadFrame) then
    begin
      FPlayer.paint;
    end;
    Suspend;
  end;
end;

procedure TFFMpegPlayer.SetBitmapHeader;
begin
  //BMP Header 
    //24bit
    with FBmpInfo.bmiHeader do
    begin
      biSize:= sizeof(TBitmapInfoHeader);
      biWidth:= FWinControl.ClientWidth;
    //Notice: BMP storage pixel data in opposite direction of Y-axis (from bottom to top).
      biHeight:= FWinControl.ClientHeight;
      biPlanes:= 1;
      biBitCount:= 24;
      biCompression:= BI_RGB;
      //biSizeImage:= ((biWidth * biBitCount+31) div 32) * 4 * biHeight;
      biSizeImage:= 0;
      biXPelsPerMeter:= 0;
      biYPelsPerMeter:= 0;
      biClrUsed:= 0;
      biClrImportant:= 0;
    end;
end;

procedure TFFMpegPlayer.CheckIndex(var startIndex: integer; var endIndex: integer);
var
  tmpIndex: integer;
begin
  if startIndex>endIndex then
  begin
    tmpIndex:= endIndex;
    endIndex:= startIndex;
    startIndex:= tmpIndex;
  end;
  if startIndex<0 then
    startIndex:= 0;
  if endIndex>=FFfms.FrameCount then
    endIndex:= FFfms.FrameCount-1;
end;

constructor TFFMpegPlayer.Create(AOwner: TWinControl);
begin
  inherited Create;
  fState:= StateNone;
  FWinControl:= AOwner;
{
  FParent:= parent;

  FWinControl:= TWinControl.Create(FParent);
  FWinControl.Name:= 'FWinControl';
  FWinControl.Parent:= FParent;
  FWinControl.Align:= alClient;
}

  FImage:= TImage.Create(FWinControl);
  FImage.Parent:= FWinControl;
  FImage.Align:= alClient;
  FImage.Stretch:= true;

  FRectTracker:= TDXRectTracker.Create(FWinControl);
  FRectTracker.Name:= 'FRectTracker';
  FRectTracker.Parent:= FWinControl;
  FRectTracker.Left:= (FWinControl.ClientWidth-FRectTracker.Width) div 2;
  FRectTracker.Top:= (FWinControl.ClientHeight-FRectTracker.Height) div 2;

  FPlayThread:= TPlaybackThread.Create(self);
  FTimer:=TMMTimer.Create;
  FTimer.OnTimeProc:= OnTimer;

//  FFrameIndex:= -1;

  FGif:=TGIFImage.Create;
  FGif.ColorReduction:= rmQuantize;

  Fffms:= TFFmpegSource.Create;
  Fffms.Width:= FWinControl.ClientWidth;
  Fffms.Height:= FWinControl.ClientHeight;

  FTimer.Internal:= 40;

  Fhdc:= GetDC(FWinControl.Handle);
//  SetStretchBltMode(Fhdc, COLORONCOLOR);
  Fhdd:= DrawDibOpen;
end;

destructor TFFMpegPlayer.Destroy;
begin
  DrawDibClose(Fhdd);
  ReleaseDC(FWinControl.Handle, Fhdc);

  FreeAndNil(FTimer);
  FPlayThread.Terminate;
  sleep(40);

  FreeAndNil(FGif);
  FreeAndNil(Fffms);
  Freeandnil(FRectTracker);
  Freeandnil(FImage);

  inherited;
end;

function TFFMpegPlayer.GetError: string;
begin
  result:= Fffms.Error;
end;

function TFFMpegPlayer.GetTime:double;
begin
  result:= Fffms.Time;
end;

function TFFMpegPlayer.GetDuration: double;
begin
  result:= Fffms.Duration;
end;

function TFFMpegPlayer.GetTimeStamp: int64;
begin
end;

function TFFMpegPlayer.GetFrameCount:integer;
begin
  result:= Fffms.FrameCount;
end;

function TFFMpegPlayer.GetVideoWidth: integer;
begin
  result:= Fffms.VideoWidth;
end;

function TFFMpegPlayer.GetVideoHeight: integer;
begin
  result:= Fffms.VideoHeight;
end;

function TFFMpegPlayer.GetFPS: double;
begin
  result:= Fffms.FPS;
end;

function TFFMpegPlayer.GetFrameIndex: integer;
begin
  result:= Fffms.FrameIndex;
end;

function TFFMpegPlayer.GetIndexTime(index:integer):double;
begin
  //result:= FFFMS.GetPTS(index) * FTimeBase;
end;

function TFFMpegPlayer.GetSelection: TSelectedRect;
var
  ratio: Double;
begin
  ratio:= 1;
  if Fffms.IsValid then
    ratio:= Fffms.VideoWidth / FWinControl.Width;
  result.Left:= trunc(FRectTracker.Left * ratio);
  result.Top:= trunc(FRectTracker.Top * ratio);
  result.Width:= trunc(FRectTracker.Width * ratio);
  result.Height:= trunc(FRectTracker.Height * ratio);
end;

procedure TFFMpegPlayer.SetFileName(value: string);
begin
  if (length(value)<=0) then
    fFileName:=''
  else
  begin
    fFileName:= value;
    OpenFile;
  end;
end;

{
procedure TFFMpegPlayer.SetTime(value: double);
begin

  if (value<0) or (value>FDuration) or (value=FTime) then
    exit;
  if (State in [StateNone, StateError]) then
    exit;
  Pause;
  if SeekTime(value) then
    Paint;
end;
}

procedure TFFMpegPlayer.SetOnStartIndex(value: TFFMpegStartIndexEvent);
begin
  FOnStartIndex:= value;
end;

procedure TFFMpegPlayer.SetOnEndIndex(value: TNotifyEvent);
begin
  FOnEndIndex:= value;
end;

procedure TFFMpegPlayer.SetOnIndexChanged(value: TFFMpegIndexChangedEvent);
begin
  FOnIndexChanged:= value;
end;

procedure TFFMpegPlayer.SetOnLoaded(value: TNotifyEvent);
begin
  FOnLoaded:= value;
end;

procedure TFFMpegPlayer.SetOnPaused(value: TNotifyEvent);
begin
  FOnPaused:= value;
end;

procedure TFFMpegPlayer.SetOnRefresh(value: TNotifyEvent);
begin
  FOnRefresh:= value;
end;

procedure TFFMpegPlayer.SetOnError(value: TNotifyEvent);
begin
  FOnError:= value;
end;

procedure TFFMpegPlayer.SetOnSelectionResize(value: TNotifyEvent);
begin
  FOnSelectionResize:= value;
  FRectTracker.OnResize:= FOnSelectionResize;
end;

procedure TFFMpegPlayer.ShowRect;
var
  tmpBmp: TBitmap;
begin
  tmpBmp:= TBitmap.Create;
  try
    Fffms.GetFrameBmp(tmpBmp, Fffms.FrameIndex);
    FImage.Picture.Assign(tmpBmp);
    FImage.Show;
  finally
    FreeAndNil(tmpBmp);
  end;
  FRectTracker.Show;
end;

procedure TFFMpegPlayer.HideRect;
begin
  FImage.Hide;
end;

procedure TFFMpegPlayer.OnTimer;
begin
  //Application.ProcessMessages;
  if (Fffms.FrameIndex>FEndFrameIndex) then
  begin
    pause;
  end
  else
  if assigned(FPlayThread) then
  begin
    FPlayThread.Resume
  end;
end;

procedure TFFMpegPlayer.DoError;
begin
  FState:= StateError;
  if assigned(FOnError) then
    FOnError(self);
end;

procedure TFFMpegPlayer.Play;
begin
  PlayAt(Fffms.FrameIndex, Fffms.FrameCount-1);
end;

procedure TFFMpegPlayer.Pause;
begin
  if state in [StateNone, Stateloading] then
    exit;
  FState:=StatePaused;
  FTImer.Enable:=false;
  sleep(10);
  ShowRect;
  if assigned(FOnPaused) then
    FOnPaused(self);
end;

procedure TFFMpegPlayer.PlayAt(startIndex, endIndex:integer);
begin
  if state in [StateNone, StateLoading, StateError] then
    exit;
  CheckIndex(startIndex, endIndex);
  FState:= StatePlaying;
  if (startIndex<>FrameIndex) then
  begin
    if not (FFFMS.SeekFrame(startIndex)) then
    begin
      DoError;
      exit;
    end;
  end;
  FEndFrameIndex:= endIndex;
  FTimer.Enable:= true;
end;

procedure TFFMpegPlayer.GoNextFrame;
begin
  ReadFrame;
  paint;
  ShowRect;
end;

procedure TFFMpegPlayer.GoPrevFrame;
begin
  if (Fffms.FrameIndex<=0) then
    exit;
  GoFrame(Fffms.FrameIndex-1);
end;

procedure TFFMpegPlayer.GoFrame(index: integer);
begin
  if FState=Stateplaying then
    Pause;
  if not Fffms.SeekFrame(index) then
  begin
    DoError;
    exit;
  end;
  FPBGRData:= Fffms.ReadBGRData;
  if (FPBGRData=nil) then
  begin
    DoError;
    exit;
  end;
  paint;
  ShowRect;
end;

function TFFMpegPlayer.ReadFrame: boolean;
begin
  result:= false;
  if (Fffms.FrameIndex>=Fffms.FrameCount-1) then
    exit;
  if not Fffms.SeekFrame(Fffms.FrameIndex+1) then
  begin
    DoError;
    exit;
  end;
  FPBGRData:= Fffms.ReadBGRData;
  result:= FPBGRData<>nil;
  if not result then
  begin
    DoError;
    exit;
  end;
end;

procedure TFFMpegPlayer.OpenFile;
begin
  if fileexists(FileName) then
  begin
    fState:=StateLoading;
    if Fffms.LoadVideo(filename) then
    begin
      FEndFrameIndex:= Fffms.FrameCount-1;
      FTimer.Internal:= round(1000 / Fffms.FPS);
      if FWinControl.ClientWidth<512 then
        FWinControl.ClientWidth:= 512;
      FWinControl.ClientHeight:= VideoHeight * FWinControl.ClientWidth div VideoWidth;

      FRectTracker.Left:= (FWinControl.ClientWidth-FRectTracker.Width) div 2;
      FRectTracker.Top:= (FWinControl.ClientHeight-FRectTracker.Height) div 2;

      Fffms.Width:= FWinControl.ClientWidth;
      Fffms.Height:= FWinControl.ClientHeight;
      Fffms.PixelFormat:= BitmapPixelFormat;

      SetBitmapHeader;

      GoFrame(0);

      if assigned(OnLoaded) then
        Onloaded(self);

      if assigned(OnSelectionResize) then
        OnSelectionResize(self);

      fState:=StateLoaded;
    end;
  end;
end;

procedure TFFMpegPlayer.Paint;
begin
  if (FPBGRData<>nil) then
  begin
    SetBitmapHeader;
    DrawDibDraw(Fhdd, Fhdc,
                0, 0, FWinControl.ClientWidth, FWinControl.ClientHeight,
                @(FBmpInfo.bmiHeader),
                FPBGRData,
                0, 0, FBmpInfo.bmiHeader.biWidth, FBmpInfo.bmiHeader.biHeight,
                DDF_SAME_HDC);
  end;
  if Assigned(FOnRefresh) then
    FOnRefresh(Self);
end;

function TFFMpegPlayer.ToGif(startIndex, endIndex, width, height, fps:integer):boolean;
var
  bmp, tmpbmp: TBitmap;
  delay: Word; // in centiseconds
  interval: integer;
  frameIndex, gifIndex:integer;
  gifExt: TGIFGraphicControlExtension;
  gifLoopExt: TGIFAppExtNSLoop;
  gifOptions: TGIFOptimizeOptions;
  selection: TSelectedRect;
begin
  result:= false;
  CheckIndex(startIndex, endIndex);

  delay:= round(100/fps);  // in centiseconds;
  interval:= round(GetFPS/fps);
  if interval<1 then
    interval:= 1;
  selection:= GetSelection;
  FGif.Clear;
  bmp:= TBitmap.Create;
  tmpbmp:= TBitmap.Create;
  frameIndex:= startIndex;
  bmp.Width:= width;
  bmp.Height:= height;
  try
    while (frameIndex<=endIndex) do
    begin
      if (frameIndex-startIndex) mod interval=0 then
      begin
        Fffms.GetFrameBmp(tmpbmp, frameIndex);
        StretchBlt(bmp.Canvas.Handle,
          0,0, width, height,
          tmpBmp.Canvas.Handle,
          selection.Left, selection.Top, selection.Width, selection.Height,
          SRCCOPY);
        gifIndex:= FGif.add(bmp);
        // Add Graphic Control Extension (for delay)
        gifExt:= TGIFGraphicControlExtension.Create(FGif.Images[gifIndex]);
        gifExt.Delay:= delay;
        FGif.Images[gifIndex].Extensions.Add(gifExt);
      end;
      inc(frameIndex);
    end;
  finally
    freeandnil(tmpbmp);
    freeandnil(bmp);
  end;
  if FGif.Images.Count>0 then
  begin
    // Add Netscape Loop extension first!
    gifLoopExt:= TGIFAppExtNSLoop.Create(FGif.Images[0]);
    gifLoopExt.Loops:= 0; // Forever
    FGif.Images[0].Extensions.Add(gifLoopExt);
    
    gifOptions:= [ooMerge, ooCrop, ooCleanup, ooColorMap];
    FGif.Optimize(gifOptions, rmNone, dmNearest, 0);
  end;
  result:= true;
end;

end.
